#! /usr/local/bin/gawk -f
### /u/sy/beebe/xml/shbook/awk/xmlfixup.awk, Wed Jul  2 13:21:39 2003
### Edit by Nelson H. F. Beebe <beebe@math.utah.edu>
### ====================================================================
### TO DO: Add code to preserve declarations and directives verbatim: 
### this should be done with a new tag class!
### ====================================================================
### Indent XML markup according to tag nesting level.  Tags listed in
### the Exclude[] table do not receive any additional indentation or
### line breaking.
###
### TAB is treated like space; XML file should normally not have TABs.
###
### Usage:
###	gawk [-v ConfigFile=file] [-v Debug=nnn] [-v DumpConfig=file] \
###	     [-v Indent=nnn] [-v Level=nnn] [-v Width=nnn] \
###	     -f xmlfixup.awk file(s)
###
###
### [08-Jul-2003]	Version 1.00
### [02-Dec-2003]	Version 1.01
### ====================================================================

BEGIN	{ initialize() }

	{ do_record($0) }

### ====================================================================


function debug_protect(s)
{
    ## Convert newlines to visible escape sequences
    gsub("\n", "\\n", s)
    return (s)
}


function do_config()
{
    if (ConfigFile == "")
	do_config_default()
    else
	do_config_file()
    if (DumpConfig != "")
	dump_config()
}


function do_config_default(    config_default,k,lines,n)
{
    ## Formatting rules indexed by XML tag

    ## This initialization is intentionally done in such a way that a
    ## configuration file dump can be inserted in this string, with
    ## space-backslash-n-backslash appended to each line.  The code in
    ## do_config_line() then handles both internal and external
    ## configuration settings.

    config_default = " \
		Begin_Verbatim    : <literallayout> <screen> \n\
 \n\
		Break_Before      : <colspec> <systemitem> <xref> \n\
 \n\
		Empty_Line_After  : </entry> </item> </listitem> </para> \n\
		Empty_Line_After  : </row> </screen> </tbody> </thead> \n\
		Empty_Line_After  : </varlistentry> \n\
 \n\
		Empty_Line_Before : <entry> <item> <listitem> <para> <row> \n\
		Empty_Line_Before : <screen> <tbody> <thead> <varlistentry> \n\
		Empty_Line_Before : <xref> \n\
 \n\
		End_Verbatim      : </literallayout> </screen> \n\
 \n\
		No_Break_Before   : <footnote> \n\
 \n\
		Ordinary          : <!> </citetitle> </command> </emphasis> \n\
		Ordinary          : </envvar> </filename> </firstterm> \n\
		Ordinary          : </indexterm> </literal> </literallayout> \n\
		Ordinary          : </option> </quote> </replaceable> \n\
		Ordinary          : </screen> </secondary> </subscript> \n\
		Ordinary          : </superscript> </systemitem> </term> \n\
		Ordinary          : </tertiary> </title> </ulink> </xref> \n\
		Ordinary          : <?> <citetitle> <colspec> <command> \n\
		Ordinary          : <emphasis> <envvar> <filename> \n\
		Ordinary          : <firstterm> <footnoteref> <indexterm> \n\
		Ordinary          : <literal> <literallayout> <option> \n\
		Ordinary          : <primary> <quote> <replaceable> <screen> \n\
		Ordinary          : <secondary> <spanspec> <subscript> \n\
		Ordinary          : <superscript> <systemitem> <term> \n\
		Ordinary          : <tertiary> <title> <ulink> <xref> \n\
    "

    ConfigFile = "internal"
    n = split(config_default, lines, "\n")
    for (k = 1; k <= n; ++k)
	do_config_line(lines[k])
}


function do_config_file(   line)
{
    while ((getline line < ConfigFile) > 0)
	do_config_line(line)
    close(ConfigFile)
}


function do_config_line(line,  fnr,parts)
{
    fnr++
    gsub("\t", " ", line)		# reduce TABs
    sub(" *#.*$", "", line)		# strip comments
    sub("^ +", "", line)		# trim leading space
    sub(" +$", "", line)		# trim trailing space

    split(line, parts, "[ \t]*:[ \t]*")
    out_debug("do_config_line(): Line = [" line "] parts[1] = [" parts[1] "] parts[2] = [" parts[2] "]")
    if (line ~ "^ *$")
	return				# ignore empty lines
    else if (parts[1] == "Begin_Verbatim")
	load_defaults("Begin_Verbatim", Begin_Verbatim, parts[2])
    else if (parts[1] == "Break_Before")
	load_defaults("Break_Before", Break_Before, parts[2])
    else if (parts[1] == "Empty_Line_After")
	load_defaults("Empty_Line_After", Empty_Line_After, parts[2])
    else if (parts[1] == "Empty_Line_Before")
	load_defaults("Empty_Line_Before", Empty_Line_Before, parts[2])
    else if (parts[1] == "End_Verbatim")
	load_defaults("End_Verbatim", End_Verbatim, parts[2])
    else if (parts[1] == "No_Break_Before")
	load_defaults("No_Break_Before", No_Break_Before, parts[2])
    else if (parts[1] == "Ordinary")
	load_defaults("Ordinary", Ordinary, parts[2])
    else
	out_debug(ConfigFile ":" fnr ":Unrecognized configuration name [" parts[1] "]")
}


function do_record(s,	comment_body, rt_is_comment)
{
    if (RT == "<!--")	# skip comments entirely
    {
	rt_is_comment = 1;

    	RS = Comment_end_pat	# slurp in comment
	getline comment_body
	RS = Tag_or_comment_pat

	RT = ("<!--" comment_body RT)
    }

    RT_Name = tagname(RT)

    if (In_Verbatim)
    {
	out_verbatim(s)
	out_verbatim(RT)
    }
    else
    {
	if (rt_is_comment)
	{
	    if (index(RT, "\n") > 0)	# heuristic: multiline comments occur between line breaks
		out_newline()
	    out_verbatim(RT)
	    out_newline()
	    return
	}
	else
	    out_normal(s)
    }

    if (RT_Name in Begin_Verbatim)
	In_Verbatim = 1
    else if (RT_Name in End_Verbatim)
    {
	In_Verbatim = 0
	if (RT_Name in Empty_Line_After)
	{
	    out_newline()
	    out_newline()
	}
    }
}


function dump_config()
{
    print "# xmlfixup version", Version, Date > DumpConfig
    print "" > DumpConfig
    dump_table("Begin_Verbatim", Begin_Verbatim)
    dump_table("Break_Before", Break_Before)
    dump_table("Empty_Line_After", Empty_Line_After)
    dump_table("Empty_Line_Before", Empty_Line_Before)
    dump_table("End_Verbatim", End_Verbatim)
    dump_table("No_Break_Before", No_Break_Before)
    dump_table("Ordinary", Ordinary)
    close(DumpConfig)
}


function dump_table(name,table,    column,i,j,key,n,swap,words)
{
    ## Make word table indexed 1..n
    n = 0
    for (key in table)
	words[++n] = key

    ## Sort word table
    for (i = 1; i <= n; i++)
    {
	for (j = i; j <= n; j++)
	{
	    if (words[i] > words[j])
	    {
		swap = words[i]
		words[i] = words[j]
		words[j] = swap
	    }
	}
    }

    column = 0
    for (i = 1; i <= n; ++i)
    {
	if ((column == 0) || ((column + 1 + length(words[i])) > Width))
	{
	    if (column > 0)
	    {
	        printf("\n") > DumpConfig
		column = 0
	    }
	    if (i == 1)
		printf("%-17s : # clear existing list\n", name) > DumpConfig
	    printf("%-17s : ", name) > DumpConfig
	    column += 20
	}
	printf("%s%s", ((column > 20) ? " " : ""), words[i]) > DumpConfig
	column += ((column > 20) ? 1 : 0) + length(words[i])
    }
    if (column > 0)
	printf("\n") > DumpConfig

    printf("\n") > DumpConfig
}


function indent_to_level(level)
{
    return (substr(BLANKS, 1, level * Indent))
}


function initialize()
{
    Line_Buffer = ""
    Version = "1.01"
    Date = "[02-Dec-2003]"

    if (Debug == "")
	Debug = 0

    if (Indent == "")
	Indent = 2			# default number of spaces per indentation level
    Indent = int(0 + Indent)		# coerce to integer
    if (Indent < 0)
	Indent = 0

    In_Verbatim = 0

    if (Level == "")			# outer indentation level
	Level = 0
    Level = int(0 + Level)		# coerce to integer
    if (Level < 0)
	Level = 0

    Level_Pending = Level

    if (Width == "")			# default output line width
	Width = 60
    Width = int(0 + Width)
    if (Width < 0)
	Width = 0
    if (Width == 0)			# zero width implies `infinite' width
	Width = 2^52 + (2^52 - 1)

    do_config()

    Tag_or_comment_pat = "(<[^!][^<>]*>|<!--)"
    Comment_end_pat = "-->"
    RS = Tag_or_comment_pat		# XML tag or comment-start separates records
    BLANKS = "                                                            "
#    BLANKS = "____________________________________________________________"

}


function is_empty_line()
{
    return (Line_Buffer ~ /^ *$/)
}


function load_defaults(name,array,values,  k,key,n,parts)
{
    sub("^ +", "", values)		# trim leading space
    sub(" +$", "", values)		# trim trailing space

    out_debug("load_defaults(\"" name "\", array, \"" values "\"")

    if (values == "")			# special case: clear all previous entries
	delete array
    else
    {
	n = split(values, parts, " ")
	for (k = 1; k <= n; k++)
	    array[parts[k]]++
	out_debug("--------------------")
	for (key in array)
	    out_debug("load_defaults(): " name "[\"" key "\"]")
    }
}


function open_tag(tag)
{
    sub("</", "<", tag)
    return (tag)
}


function out_debug(s)
{
    if (Debug)
	print "DEBUG:", s > "/dev/stderr"
}


function out_newline()
{
    out_verbatim("\n")
}


function out_normal(s,  eol,indent,tag_indent)
{
    out_debug("------------------------------------------------------------------------")
    out_debug("out_normal(" s ")  Indent = " Indent \
	      " Level = "  Level \
	      " Level_Pending = " Level_Pending \
	      " Column = " Column \
	      " Width = " Width)
    out_debug("s = [" s "]\tRT = [" RT "]")

    gsub("\n",  " ", s)			# collapse lines
    sub("^ +", " ", s)			# reduce leading space
    sub(" +$", " ", s)			# reduce trailing space
    gsub("  +", " ", s)			# reduce embedded space
    gsub("[?!:.][ ]", "& ", s)		# require double space after certain punctuation

    gsub("\n",  " ", RT)		# collapse lines
    gsub("  +", " ", RT)		# reduce embedded space

    if (s ~ /^ *$/)
	s = ""

    if (Level_Pending > 0)
    {
	indent = indent_to_level(Level_Pending)
	Level_Pending = 0
    }
    else if (Column == 0)
	indent = indent_to_level(Level)

    if ((Column >= Width) && (s ~ /^[ ]/))	# Line wrap okay only if space follows last tag
    {
	out_newline()
	indent = indent_to_level(Level)
    }

    if (Column == 0)    		# remove leading space in string at start of line
	sub(/^ +/,"",s)
    out_verbatim(indent)
    out_wrap(s)

    if (RT_Name in Empty_Line_Before)
    {
	if (!is_empty_line())	# nonempty previous line
	    out_newline()
	out_newline()
    }

    if (RT_Name in Break_Before)
    {
	if (Column > 0)
	    out_newline()
	tag_indent = indent_to_level(Level)
	eol = ""
    }
    else if (RT_Name in Ordinary)
    {
	tag_indent = ""			# no extra indentation required
	eol = ""
    }
    else if (RT ~ /<\//)		# standalone close tag
    {
	Level--
	if (Level < 0) 			# sanity check
	    Level = 0
	if (Tag_Stack[Level] != open_tag(RT))
	    warning("Close tag " RT " does not match open tag " Tag_Stack[Level])
	if (Column > 0)			# nonempty previous line
	    out_newline()
	tag_indent = indent_to_level(Level)
	Level_Pending = Level
	eol = "\n"
    }
    else				# standalone open tag
    {
	if (RT_Name in No_Break_Before)
	    tag_indent = ""
	else
	{
	    if (Column > 0)		# nonempty previous line
		out_newline()
	    tag_indent = indent_to_level(Level)
	}
	Tag_Stack[Level] = RT_Name
	Level++
	Level_Pending = Level
	eol = "\n"
    }

    out_verbatim(tag_indent)
    out_verbatim(RT)
    out_verbatim(eol)

    if (RT_Name in Empty_Line_After)
	out_newline()
}


function out_verbatim(s, n,parts)
{
    out_debug("begin out_verbatim(" s ")  length(s) = " length(s) \
	      "  Column = " Column "  RT_Name = [" RT_Name "]" \
	      "  Newline_Count = " Newline_Count)

    out_debug("  Line_Buffer = [" debug_protect(Line_Buffer) "]")
    if (s == "")			# ignore empty strings to preserve Newline_Count
	return
    else if (s == "\n")
    {
	Newline_Count++
	if (Newline_Count >= 3)	# prevent more than one successive empty line
	    return
    }
    else
	Newline_Count = 0
    if (s == "\n")
    {
	sub(" +$", "", Line_Buffer)	# trim trailing space
	printf("%s\n", Line_Buffer)	# all program output happens here
	Newline_Count++
	Line_Buffer = ""
	Column = 0
    }
    else
    {
	Line_Buffer = Line_Buffer s
	n = split(s,parts,"\n")		# s is often multiline, so split needed
	Column = (n <= 1) ? (Column + length(s)) : length(parts[n])
    }
    out_debug("end out_verbatim(" s ")  Column = " Column "   parts[" n "] = [" parts[n] "]  RT_Name = [" RT_Name "]")
    out_debug("")
}


function out_wrap(s, k,n,words)
{
    out_debug("")
    out_debug("out_wrap(" s ") Column = " Column)
    if ((Column + length(s)) <= Width)
	out_verbatim(s)
    else
    {
	n = split(s,words,"[ ]")	# NB: must be single space, to avoid loss of space between text and <tag>
	if (n <= 1)			# long token producing line longer than Width
	    out_verbatim(s)
	else
	{
	    for (k = 1; k <= n; ++k)
	    {
		out_debug("out_wrap(): words[" k "] = [" words[k] "]   Column = " Column)
		if ((Column + length(words[k])) <= Width)
		{
		    if ((k > 1) && !is_empty_line())	# at least one nonblank token on line
			out_verbatim(" ") # restore word separator
		    out_verbatim(words[k])
		}
		else			# time to wrap line
		{
		    if (k > 1)
		    {
		        out_newline()
		        out_verbatim(indent_to_level(Level))
		    }
		    out_verbatim(words[k])
		}
	    }
	}
    }
}


function tagname(s)
{
    sub("<!.*>", "<!>", s)		# special case for comments and entity definitions
    sub("<[?][-a-z]+.*>", "<?>", s)	# special case for directives
    sub(" .*>", ">", s)			# reduce <sect1 id="..."> to <sect1>, etc.
    return (s)
}


function warning(s)
{
    print "WARNING: " FILENAME ":" s > "/dev/stderr"
    gsub("[&]", "\\&amp;", s)
    gsub("<", "[", s)
    gsub(">", "]", s)
    print "\n<!-- WARNING: " s " -->\n"
}
